package com.aware.plugin.beacons;

import android.bluetooth.*;
import android.content.*;
import android.net.Uri;
import android.os.Handler;
import android.preference.PreferenceManager;
import android.util.Log;
import com.aware.Aware;
import com.aware.Aware_Preferences;
import com.aware.plugin.beacons.db.BeaconScanBroadcastReceiver;
import com.aware.plugin.beacons.db.Beacons_Provider;
import com.aware.plugin.beacons.model.*;
import com.aware.plugin.beacons.parser.NotiBeaconException;
import com.aware.plugin.beacons.parser.iBeaconPacketParser;
import com.aware.plugin.beacons.view.BluetoothRequestActivity;
import com.aware.utils.Aware_Sensor;
import com.kontakt.sdk.android.device.Beacon;

import java.util.Random;
import java.util.UUID;

import static com.aware.plugin.beacons.Constants.*;

public class Plugin extends Aware_Sensor {

    private boolean DEBUG = false;
    private boolean mockBeacons = false;
    private Handler handler = new Handler();

    private BluetoothAdapter mBluetoothAdapter;
    private BroadcastReceiver scanReceiver = new BeaconScanBroadcastReceiver();

    @Override
    public void onCreate() {
        super.onCreate();

        //Set plugin values on the framework
        TAG = Constants.BEACONS_PLUGIN_TAG;
        DEBUG = Aware.getSetting(this.getContentResolver(), Aware_Preferences.DEBUG_FLAG).equals("true");

        //Share the context back to the framework and other applications
        CONTEXT_PRODUCER = new Aware_Sensor.ContextProducer() {
            @Override
            public void onContext() {
                Intent notification = new Intent(Constants.ACTION_AWARE_BEACONS);
                sendBroadcast(notification);
            }
        };

        //Remote database replication
        DATABASE_TABLES = Beacons_Provider.DATABASE_TABLES;
        TABLES_FIELDS = Beacons_Provider.TABLES_FIELDS;
        CONTEXT_URIS = new Uri[]{ Beacons_Provider.Beacon.CONTENT_URI,
                Beacons_Provider.Experiment.CONTENT_URI,
                Beacons_Provider.Calibration_Readings_Base.CONTENT_URI,
                Beacons_Provider.Calibration_Readings.CONTENT_URI};

        final BluetoothManager bluetoothManager =
                (BluetoothManager) getSystemService(Context.BLUETOOTH_SERVICE);
        mBluetoothAdapter = bluetoothManager.getAdapter();

        if (mBluetoothAdapter == null || !mBluetoothAdapter.isEnabled()) {
            //Bluetooth not turned on, request
            Intent dialogIntent = new Intent(this, BluetoothRequestActivity.class);
            dialogIntent.addFlags(Intent.FLAG_ACTIVITY_NEW_TASK);
            if(DEBUG){
                Log.d(Constants.BEACONS_PLUGIN_TAG, "Starting bluetooth request activity");
            }
            this.startActivity(dialogIntent);
        }else{
            if(DEBUG){
                Log.d(Constants.BEACONS_PLUGIN_TAG, "Bluetooth turned on already, start scan");
            }
            startScan();
        }

        registerReceiver(bluetoothStateReceiver, new IntentFilter(BluetoothAdapter.ACTION_STATE_CHANGED));
        registerReceiver(scanReceiver, new IntentFilter(SCAN_RESULT_ACTION));

        if(DEBUG){
            Log.d(Constants.BEACONS_PLUGIN_TAG, "Beacons created");
        }
    }

    @Override
    public void onDestroy() {
        stopScan();
        unregisterReceiver(scanReceiver);
        unregisterReceiver(bluetoothStateReceiver);
        if(DEBUG){
            Log.d(Constants.BEACONS_PLUGIN_TAG, "Beacons destroyed");
        }
        super.onDestroy();
    }

    private void startScan(){
        try {
            mBluetoothAdapter.startLeScan(mLeScanCallback);
        }
        catch (NullPointerException e) {
            Log.w(Constants.BEACONS_PLUGIN_TAG, "Bluetooth adapter is null!");
        }
        handler.postDelayed(new Runnable() {
            @Override
            public void run() {
                stopScan();
            }
        }, 350);
    }

    private void stopScan() {
        try {
            mBluetoothAdapter.stopLeScan(mLeScanCallback);
        } catch (NullPointerException e) {
            Log.w(Constants.BEACONS_PLUGIN_TAG, "Bluetooth adapter is null!");
        }
        handler.post(new Runnable() {
            @Override
            public void run() {
                startScan();
            }
        });
    }

    private BroadcastReceiver bluetoothStateReceiver = new BroadcastReceiver() {
        @Override
        public void onReceive(Context context, Intent intent) {
            final String action = intent.getAction();

            if (action != null && action.equals(BluetoothAdapter.ACTION_STATE_CHANGED)) {
                final int state = intent.getIntExtra(BluetoothAdapter.EXTRA_STATE,
                        BluetoothAdapter.ERROR);
                switch (state) {
                    case BluetoothAdapter.STATE_OFF:
                        if(DEBUG){
                            Log.d(Constants.BEACONS_PLUGIN_TAG, "Bluetooth turned off, stop scan");
                        }
                        stopScan();
                        break;

                    case BluetoothAdapter.STATE_ON:
                        if(DEBUG){
                            Log.d(Constants.BEACONS_PLUGIN_TAG, "Bluetooth turned on, start scan");
                        }
                        startScan();
                        break;
                }
            }
        }
    };

    private BluetoothAdapter.LeScanCallback mLeScanCallback = new BluetoothAdapter.LeScanCallback() {
        @Override
        public void onLeScan(final BluetoothDevice device, final int rssi,
                             final byte[] scanRecord) {

            try {
                //rssi & 0xFF
                ScanResult result = iBeaconPacketParser.parse(scanRecord, rssi);

                AbstractBeacon scanBeacon;

                if(device.getName() != null && device.getName().equals("Kontakt")) {
                    scanBeacon = new KontaktBeacon(new Beacon(device, rssi, scanRecord), result);
                }else{
                    scanBeacon = new OtherBeacon(device, rssi, result);
                }

                AbstractBeacon existingBeacon;
                if((existingBeacon = Singleton.getInstance()
                        .getExistingBeacon(scanBeacon)) != null){

                    if(scanBeacon instanceof KontaktBeacon) {
                        ((KontaktBeacon) existingBeacon)
                                .setBeacon(((KontaktBeacon) scanBeacon).getBeacon());
                    }else{
                        ((OtherBeacon)existingBeacon).setBLEScanData(device, rssi, result);
                    }
                    existingBeacon.setScanResult(result);
                    scanBeacon = existingBeacon;

                }else{
                    SharedPreferences sharedPref = PreferenceManager.getDefaultSharedPreferences(Plugin.this);

                    scanBeacon.setLogCurveParameter_A(
                            Double.valueOf(sharedPref.getString(LOG_CURVE_PARAMETER_A, "-65")));
                    scanBeacon.setLogCurveParameter_b(
                            Double.valueOf(sharedPref.getString(LOG_CURVE_PARAMETER_B, "6")));
                    scanBeacon.setLogCurveParameter_n(
                            Double.valueOf(sharedPref.getString(LOG_CURVE_PARAMETER_N, "1.5")));

                    Singleton.getInstance().addFoundDevice(scanBeacon);
                }

                // d = b^((A-RSSI)/(10*n))
                scanBeacon.setDistance(
                        Math.pow(scanBeacon.getLogCurveParameter_b(),
                                ((scanBeacon.getLogCurveParameter_A() - rssi) /
                                        (10 * scanBeacon.getLogCurveParameter_n()))
                        ));

                Intent intent = new Intent(SCAN_RESULT_ACTION);
                intent.putExtra(BEACON_EXTRA, scanBeacon);
                sendBroadcast(intent);

                if(mockBeacons){
                    sendMockBeaconsData();
                }

            } catch (NotiBeaconException e) {
                //info ze znaleziono beacona, ale niespelniajacgo ibeacon
                //tymczasowo wylaczone bo zasmieca log
//                Log.w(Constants.BEACONS_PLUGIN_TAG, "Beacon " + device.getAddress() + " is not an iBeacon");
            }
        }
    };

    private void sendMockBeaconsData(){
        MockBeacon mockBeacon1 = new MockBeacon(
                (int) ( -1 * ((new Random().nextFloat() * 40) + 40)),
                "69:34:BF:40:0A:12", "MockBeacon1",
                new UUID(0xa10f8a20717511e4l, 0x82f80800200c9a66l),
                1234, 1001, -65, 1, -65, 6, 1.5);
        sendBroadcast(new Intent(SCAN_RESULT_ACTION).putExtra(BEACON_EXTRA, mockBeacon1));
        Singleton.getInstance().addFoundDevice(mockBeacon1);

        MockBeacon mockBeacon2 = new MockBeacon(
                (int) ( -1 * ((new Random().nextFloat() * 40) + 40)),
                "14:00:76:37:FC:EF", "MockBeacon2",
                new UUID(0x2ba01ab0717611e4l, 0x82f80800200c9a66l),
                1234,  1002, -65, 1, -65, 6, 1.5);
        sendBroadcast(new Intent(SCAN_RESULT_ACTION).putExtra(BEACON_EXTRA, mockBeacon2));
        Singleton.getInstance().addFoundDevice(mockBeacon2);

        MockBeacon mockBeacon3 = new MockBeacon(
                (int) ( -1 * ((new Random().nextFloat() * 40) + 40)),
                "0B:21:DE:EA:B8:BA", "MockBeacon3",
                new UUID(0x9d6769b0717511e4l, 0x82f80800200c9a66l),
                1234, 1003, -65, 1, -65, 6, 1.5);
        sendBroadcast(new Intent(SCAN_RESULT_ACTION).putExtra(BEACON_EXTRA, mockBeacon3));
        Singleton.getInstance().addFoundDevice(mockBeacon3);
    }
}
