package com.aware.plugin.beacons.micromap;

import java.io.ByteArrayOutputStream;
import java.io.IOException;
import java.io.InputStream;
import java.util.LinkedList;

import javax.xml.parsers.DocumentBuilder;
import javax.xml.parsers.DocumentBuilderFactory;
import javax.xml.parsers.ParserConfigurationException;
import javax.xml.transform.Transformer;
import javax.xml.transform.TransformerFactory;
import javax.xml.transform.dom.DOMSource;
import javax.xml.transform.stream.StreamResult;

import android.content.Context;
import android.content.res.AssetManager;
import org.w3c.dom.Document;
import org.w3c.dom.Element;
import org.w3c.dom.Node;
import org.w3c.dom.NodeList;
import org.xml.sax.SAXException;

import com.vividsolutions.jts.geom.GeometryFactory;
import com.vividsolutions.jts.geom.Point;
import com.vividsolutions.jts.geom.Polygon;
import com.vividsolutions.jts.io.gml2.GMLReader;
import pl.edu.agh.geist.micromaps.Area;
import pl.edu.agh.geist.micromaps.Beacon;
import pl.edu.agh.geist.micromaps.MicroMap;
import pl.edu.agh.geist.micromaps.Obstacle;

public class GMLParser {

	public static String SUPPORTED_VERSION = "1.0";
	/*
	 * The Micro Map configuration file tags
	 */
	public static final String MMC_ROOT = "configuration";
	public static final String MMC_VERSION = "version";
	public static final String MMC_AREA = "area";
	public static final String MMC_OBSTACLES = "obstacles";
	public static final String MMC_BEACONS = "beacons";
	public static final String MMC_SRC = "src";

	/*
	 * The remaining GML tags
	 */
	public static final String GML_FEATURE = "Feature";
	public static final String GML_LABEL = "label";
	public static final String GML_POLYGON = "gml:Polygon";
	public static final String GML_PROPERTY = "property";
	public static final String GML_NAME = "name";
	public static final String GML_POINT = "gml:Point";
	public static final String GML_UUID = "uuid";
	public static final String GML_MINOR = "minor";
	public static final String GML_MAJOR = "major";

	public static MicroMap parseMicroMap(String configurationURI, Context context) throws ParserConfigurationException, SAXException, IOException{

		MicroMap map = new MicroMap();
		LinkedList<ConfigurationElement> configurationElements = null;

		AssetManager assetManager = context.getAssets();
		InputStream configurationStream = assetManager.open(configurationURI);

		DocumentBuilderFactory dbFactory = DocumentBuilderFactory.newInstance();
		DocumentBuilder dBuilder;

		dBuilder = dbFactory.newDocumentBuilder();
		Document configuration = dBuilder.parse(configurationStream);

		//Start parsing
		NodeList mmcNodes =  configuration.getElementsByTagName(MMC_ROOT);
		for(int index = 0; index < mmcNodes.getLength(); index++){
			Node mainNode  = mmcNodes.item(index);

			if (mainNode.getNodeType() == Node.ELEMENT_NODE) {
				Element eElement = (Element) mainNode;
				String version  = eElement.getAttribute(MMC_VERSION);
				if(version.compareToIgnoreCase(SUPPORTED_VERSION ) <0 ){
					throw new IOException("Unsupporteed configuration version");
				}
				configurationElements = parseConfiguration(eElement);
			}
		}
		configurationStream.close();

		//Start parsing areas and their elements configurations
		map.setAreas(parseAreas(configurationURI.replace("conf.mmc", ""), configurationElements, context));
		return map;
	}

	private static LinkedList<Area> parseAreas(String folderPath, LinkedList<ConfigurationElement> ce, Context context) throws ParserConfigurationException, SAXException, IOException{
		LinkedList<Area> areas = new LinkedList<>();
		for(ConfigurationElement e : ce){
			Area a = parseArea(folderPath + e.areaSrc, context);
			LinkedList<Obstacle> o = parseObstacles(folderPath + e.obstaclesSrc, context);
			LinkedList<Beacon> b = parseBeacons(folderPath + e.beaconsSrc, context);

			a.setOobstacles(o);
			a.setBeacons(b);

			areas.add(a);
		}
		return areas;
	}

	private static LinkedList<Beacon> parseBeacons(String beaconsSrc, Context context) throws SAXException, IOException, ParserConfigurationException {

		AssetManager assetManager = context.getAssets();
		InputStream obstacleStream = assetManager.open(beaconsSrc);

		GMLReader gml = new GMLReader();

		LinkedList<Beacon> beacons = new LinkedList<>();

		DocumentBuilderFactory dbFactory = DocumentBuilderFactory.newInstance();
		DocumentBuilder dBuilder;

		dBuilder = dbFactory.newDocumentBuilder();
		Document areaDoc = dBuilder.parse(obstacleStream);

		NodeList beaconsNodes =  areaDoc.getElementsByTagName(GML_FEATURE);

		for(int index = 0; index < beaconsNodes.getLength(); index++){
			Node mainNode  = beaconsNodes.item(index);
			Point beaconPoint = null;
			String beaconUUID = null;
			String minor = null;
			String major = null;

			if (mainNode.getNodeType() == Node.ELEMENT_NODE) {
				Element featureElement = (Element) mainNode;
				//retrieve polygon
				NodeList points = featureElement.getElementsByTagName(GML_POINT);

				for (int pointIndex = 0; pointIndex < points.getLength(); pointIndex++) {
					Node point = points.item(pointIndex);
					if (point.getNodeType() == Node.ELEMENT_NODE) {
						Element pointElement = (Element) point;
						String pointString = elementToString(pointElement);
						beaconPoint = (Point) gml.read(pointString, new GeometryFactory());
					}
				}
				//Only if we are in the real point feature proceed to retrieve the properties
				if (beaconPoint == null) continue;

				//retrieve property
				NodeList properties = featureElement.getElementsByTagName(GML_PROPERTY);
				for (int propIndex = 0; propIndex < properties.getLength(); propIndex++) {
					Node prop = properties.item(propIndex);
					if (prop.getNodeType() == Node.ELEMENT_NODE) {
						Element propElement = (Element) prop;
						String attributeName = propElement.getAttribute(GML_NAME);
						if (attributeName.equalsIgnoreCase(GML_UUID)) {
							beaconUUID = prop.getTextContent();
						} else if (attributeName.equalsIgnoreCase(GML_MINOR)) {
							minor = prop.getTextContent();
						} else if (attributeName.equalsIgnoreCase(GML_MAJOR)) {
							major = prop.getTextContent();
						}
					}
				}
				//create beacon
				beacons.add(new Beacon(beaconUUID, minor, major, beaconPoint.getCoordinateSequence(), beaconPoint.getFactory()));
			}
		}
		return beacons;
	}

	private static LinkedList<Obstacle> parseObstacles(String obstaclesSrc, Context context) throws ParserConfigurationException, SAXException, IOException {

		AssetManager assetManager = context.getAssets();
		InputStream obstacleStream = assetManager.open(obstaclesSrc);

		GMLReader gml = new GMLReader();

		LinkedList<Obstacle> obstacles = new LinkedList<>();

		DocumentBuilderFactory dbFactory = DocumentBuilderFactory.newInstance();
		DocumentBuilder dBuilder;

		dBuilder = dbFactory.newDocumentBuilder();
		Document areaDoc = dBuilder.parse(obstacleStream);

		NodeList obstacleNodes =  areaDoc.getElementsByTagName(GML_FEATURE);

		for(int index = 0; index < obstacleNodes.getLength(); index++){
			Node mainNode  = obstacleNodes.item(index);
			Polygon obstaclePolygon = null;
			String obstacleLabel = null;

			if (mainNode.getNodeType() == Node.ELEMENT_NODE) {
				Element featureElement = (Element) mainNode;
				//retrieve polygon
				NodeList polygons = featureElement.getElementsByTagName(GML_POLYGON);

				for(int polyIndex = 0; polyIndex < polygons.getLength(); polyIndex++){
					Node poly  = polygons.item(polyIndex);
					if (poly.getNodeType() == Node.ELEMENT_NODE) {
						Element polyElement = (Element) poly;
						String polyString = elementToString(polyElement);
						obstaclePolygon = (Polygon)gml.read(polyString, new GeometryFactory());
					}
				}

				//retrieve property
				NodeList properties = featureElement.getElementsByTagName(GML_PROPERTY);
				for(int propIndex = 0; propIndex < properties.getLength(); propIndex++){
					Node prop  = properties.item(propIndex);
					if (prop.getNodeType() == Node.ELEMENT_NODE) {
						Element propElement  = (Element) prop;
						String attributeName = propElement.getAttribute(GML_NAME);
						if(attributeName.equalsIgnoreCase(GML_LABEL)){
							obstacleLabel = prop.getTextContent();
						}
					}
				}
				//create and add obstacle only if poly is not null
				if(obstaclePolygon != null)
					obstacles.add(Obstacle.buildObstaceFromPoly(obstacleLabel, obstaclePolygon));
			}
		}
		return obstacles;
	}

	private static Area parseArea(String areaSrc, Context context) throws ParserConfigurationException, SAXException, IOException {

		AssetManager assetManager = context.getAssets();
		InputStream areaStream = assetManager.open(areaSrc);

		GMLReader gml = new GMLReader();

		DocumentBuilderFactory dbFactory = DocumentBuilderFactory.newInstance();
		DocumentBuilder dBuilder;

		dBuilder = dbFactory.newDocumentBuilder();
		Document areaDoc = dBuilder.parse(areaStream);

		NodeList areaNodes =  areaDoc.getElementsByTagName(GML_FEATURE);

		Polygon areaPolygon = null;
		String areaLabel = null;

		for(int index = 0; index < areaNodes.getLength(); index++){
			Node mainNode  = areaNodes.item(index);

			if (mainNode.getNodeType() == Node.ELEMENT_NODE) {
				Element featureElement = (Element) mainNode;
				//retrieve polygon
				NodeList polygons = featureElement.getElementsByTagName(GML_POLYGON);

				for(int polyIndex = 0; polyIndex < polygons.getLength(); polyIndex++){
					Node poly  = polygons.item(polyIndex);
					if (poly.getNodeType() == Node.ELEMENT_NODE) {
						Element polyElement = (Element) poly;
						String polyString = elementToString(polyElement);
						areaPolygon = (Polygon)gml.read(polyString, new GeometryFactory());
					}
				}

				//retrieve property
				NodeList properties = featureElement.getElementsByTagName(GML_PROPERTY);
				for(int propIndex = 0; propIndex < properties.getLength(); propIndex++){
					Node prop  = properties.item(propIndex);
					if (prop.getNodeType() == Node.ELEMENT_NODE) {
						Element propElement  = (Element) prop;
						String attributeName = propElement.getAttribute(GML_NAME);
						if(attributeName.equalsIgnoreCase(GML_LABEL)){
							areaLabel = prop.getTextContent();
						}
					}
				}
			}
		}
		if(areaPolygon != null){
			return Area.buildAreaFromPoly(areaLabel, areaPolygon);
		}else{
			return null;
		}
	}

	private static LinkedList<ConfigurationElement> parseConfiguration(Element mmc) {
		LinkedList<ConfigurationElement> configurations = new LinkedList<>();

		NodeList mmcNodes =  mmc.getElementsByTagName(MMC_AREA);

		for(int index = 0; index < mmcNodes.getLength(); index++){
			Node mainNode  = mmcNodes.item(index);

			if (mainNode.getNodeType() == Node.ELEMENT_NODE) {
				Element eElement = (Element) mainNode;
				String areaSrc  = eElement.getAttribute(MMC_SRC);
				String beaconsSrc = null;
				String obstaclesSrc = null;

				NodeList obstacles =  eElement.getElementsByTagName(MMC_OBSTACLES);
				for(int obstaclesIndex = 0; obstaclesIndex < obstacles.getLength(); obstaclesIndex++){
					Node obstacleNode  = obstacles.item(obstaclesIndex);
					if (obstacleNode.getNodeType() == Node.ELEMENT_NODE) {
						Element obstacleElement = (Element) obstacleNode;
						obstaclesSrc  = obstacleElement.getAttribute(MMC_SRC);
					}
				}

				NodeList beacons =  eElement.getElementsByTagName(MMC_BEACONS);
				for(int beaconsIndex = 0; beaconsIndex < beacons.getLength(); beaconsIndex++){
					Node beaconNode  = beacons.item(beaconsIndex);
					if (beaconNode.getNodeType() == Node.ELEMENT_NODE) {
						Element beaconElement = (Element) beaconNode;
						beaconsSrc  = beaconElement.getAttribute(MMC_SRC);
					}
				}
				configurations.add( new ConfigurationElement(areaSrc,obstaclesSrc, beaconsSrc) );
			}
		}
		return configurations;
	}

	private static String elementToString(Element element) {
		try {
			ByteArrayOutputStream baos = new ByteArrayOutputStream();
			DOMSource source = new DOMSource(element);
			StreamResult result = new StreamResult(baos);
			TransformerFactory transFactory = TransformerFactory.newInstance();
			Transformer transformer = transFactory.newTransformer();
			transformer.transform(source, result);
			return new String(baos.toByteArray());
		} catch (Exception ex) {
			ex.printStackTrace();
		}
		return null;
	}

	private static class ConfigurationElement{

		ConfigurationElement(String areaSrc, String obstaclesSrc,
							 String beaconsSrc) {
			this.areaSrc = areaSrc;
			this.obstaclesSrc = obstaclesSrc;
			this.beaconsSrc = beaconsSrc;
		}

		String areaSrc;
		String obstaclesSrc;
		String beaconsSrc;
	}
}